import pathlib
import click
from hakisto.click import hakisto_severity, hakisto_short_trace, hakisto_process_all, hakisto_file


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@hakisto_severity()
@hakisto_file(default=False)
@hakisto_short_trace()
@click.pass_context
def main(ctx, **kwargs):
    """Imuthes Utilities"""
    hakisto_process_all(**kwargs)
    ctx.ensure_object(dict)


@main.command()
@click.option("-t", "--target", type=click.Path(exists=True, path_type=pathlib.Path), required=True, help="Target")
@click.option("-l", "--link", type=click.Path(path_type=pathlib.Path), required=True, help="will be created")
@click.pass_context
def make_link(ctx, target, link):
    """Create symbolic link (or junction)"""

    make_link(target=target, link=link)



import collections.abc
from abc import ABCMeta, abstractmethod
from types import MappingProxyType

from hakisto import Logger

logger = Logger("imuthes")


# noinspection PyUnresolvedReferences
class Mapping(collections.abc.Mapping, metaclass=ABCMeta):
    """Abstract implementation using ``self._data`` for storage.

    Implement ``__init__`` to populate ``self._data``.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        raise NotImplementedError

    def __getitem__(self, item):
        return self._data[item]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return f"{str(self._data)}"


class NamedMapping(Mapping):
    """Mapping providing access via attribute name.

    If attribute name contains '_' and is not found, search is repeated with '_' replaced by '-'.

    If not found search is repeated case-insensitive.

    :param kwargs: The values to use for Mapping
    :type kwargs: dict
    """

    __slots__ = ("_data", "_normalized")

    def __init__(self, **kwargs):
        self._data = MappingProxyType({k: v for k, v in kwargs.items()})
        self._normalized = {}
        for k in self._data:
            k_lower = k.lower()
            if k_lower in self._normalized:
                logger.warning(f"NamedMapping: Duplicate key when normalizing »{k}«.", force_location=True)
            self._normalized[k.lower()] = self._data[k]

    def __search(self, item):
        if item not in self._data:
            if "_" in item:
                return self.__search(item.replace("_", "-"))
            elif not item.islower():
                return self.__search(item.lower())
            else:
                raise AttributeError(item)
        return self._data[item]

    def __getattr__(self, item):
        return self.__search(item)

    def __str__(self):
        return f"{str(self._data)}"



import pathlib
from collections import namedtuple
from typing import Iterable


def container_to_tuple(container):
    if not hasattr(container, "__contains__"):
        raise TypeError
    values = []
    if not hasattr(container, "get"):
        contents = list(container)
        return_func = lambda x: tuple(x)  # noqa: E731
    else:
        keys = list(container.keys())
        Node = namedtuple("Node", keys)
        contents = [container[k] for k in keys]
        return_func = lambda x: Node(*x)  # noqa: E731
    for item in contents:
        if not hasattr(item, "__contains__") or hasattr(item, "capitalize"):  # not a container or a string
            values.append(item)
        else:
            values.append(container_to_tuple(item))
    return return_func(values)


def parse_configuration(lines: Iterable[str]):
    result = {}
    for line in lines:
        line = line.strip()
        if not line or line.startswith("#") or "=" not in line:
            continue
        k, v = line.split("=", maxsplit=1)
        k, v = k.strip(), v.strip()
        keys = k.split(".")
        r = result
        for i in range(len(keys) - 1):
            if keys[i] not in r:
                r[keys[i]] = {}
            r = r[keys[i]]
        r[keys[-1]] = v

    return container_to_tuple(result)


def get_configuration(path: pathlib.Path):
    """Read PATH and make contents available as (nested) NamedTuple.

    :param path: File with configuration
    """
    with path.open("r") as f:
        return parse_configuration(f.readlines())



from typing import Any, Callable, Generic, TypeVar

T = TypeVar("T")


# noinspection PyPep8Naming
class class_property(property, Generic[T]):
    """Decorator for a Class-level property.

    Credit to Denis Rhyzhkov on Stackoverflow: https://stackoverflow.com/a/13624858/1280629
    """

    def __init__(self, func: Callable[[Any], T]) -> None:
        super().__init__(func)

    def __get__(self, owner_self: object, owner_cls: type | None = None) -> T:
        return self.fget(owner_cls)



# ------------------------------------------------------------------------------
#  Imuthes by NetLink Consulting GmbH
#
#  Copyright (c) 2025. Bernhard W. Radermacher
#
#  This program is free software: you can redistribute it and/or modify it under
#  the terms of the GNU Lesser General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  This program is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
#  details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

import inspect
import threading

from hakisto import Logger

logger = Logger("imuthes")
Logger.register_excluded_source_file(inspect.currentframe().f_code.co_filename)


class Singleton:
    """Thread safe singleton class.

    Just add as a super-class.
    """

    _instance = None
    _lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        with cls._lock:
            if cls._instance is None:
                logger.debug(f"Creating Singleton »{cls.__name__}«.")
                cls._instance = super(Singleton, cls).__new__(cls)
            else:
                logger.debug(f"Using Singleton »{cls.__name__}«.")
        return cls._instance



import datetime
from typing import Iterable

from hakisto import logger

from typing import Callable


def attr_getter_factory(name: str) -> Callable:
    if "." in name:
        remainder, inner_name = name.rsplit(".", 1)

        def f(obj):
            return getattr(attr_getter_factory(remainder)(obj), inner_name)

    else:

        def f(obj):
            return getattr(obj, name, "")

    return f


class DisplayColumn:
    """Utility to specify column characteristics for :class:`imuthes.DisplayTable`.

    While the column name can be sufficient, this allows to define alignment and heading of a column.

    .. list-table:: Possible values for ``format``

       * - ``<``
         - left aligned (default)
       * - ``>``
         - right aligned
       * - ``^``
         - centered


    :param name: Column name. If **class-instances** will be added to a table, this can use *dotted* format to specify deeper structure.
    :param format: For now, just the alignment.
    :param header: Header in table column.
    :param is_key: Mark as KEY column.
    :param show_none: Show None (or NULL) explicitly.
    """

    def __init__(self, name: str, format: str = "", header: str = None, is_key: bool = False, show_none: bool = False):
        self.name = name
        self.format = format
        self.header = name.title() if header is None else header
        self.is_key = is_key
        self.show_none = show_none


class DisplayTable:
    """Table to be displayed in console or browser or as Markdown.

    Values are converted to ``str`` when they are added. If you want to support your own type, make sure to add a ``__str__`` method to the class.

    ============  =================
    Type          Format
    ============  =================
    ``date``      YY-MM-DD
    ``time``      HH:MM:SS
    ``datetime``  YY-MM-DD HH:MM:SS
    ============  =================

    :param args: The columns to display. Either the name of the column, or :class:`imuthes.DisplayColumn`.
    """

    def __init__(self, *args: str | DisplayColumn | Iterable) -> None:
        self._rows = []
        self._columns = []
        for arg in args:
            if not isinstance(arg, DisplayColumn):
                arg = DisplayColumn(arg)
            self._columns.append(arg)

    def append(self, value):
        """Append row to table.

        The value can either be a dictionary with the column names, or an object that exposes
        the column names as attributes."""
        if hasattr(value, "__getitem__"):
            f = lambda x: value.get(x.name, "")  # noqa: E731
        else:  # expecting to get an object, but here the name might be nested...
            f = lambda x: attr_getter_factory(x.name)(value)  # noqa: E731
        # row = [f(i) if f(i) is not None else "_None_" for i in self._columns]
        row = [f(i) for i in self._columns]
        for n, i in enumerate(row):
            # if isinstance(i, str) and "\n" in i:
            #     row[n] = i.replace("\n", "<br/>")
            if i is None or isinstance(i, bool):
                continue
            elif isinstance(i, datetime.datetime):
                row[n] = i.strftime("%y-%m-%d %H:%M:%S")
            elif isinstance(i, datetime.date):
                row[n] = i.strftime("%y-%m-%d")
            elif isinstance(i, datetime.time):
                row[n] = i.strftime("%H:%M:%S")
            else:
                row[n] = str(i)
        self._rows.append(row)

    def __len__(self):
        return len(self._rows)

    def __getitem__(self, index):
        return self._rows[index]

    def ascii(self) -> str:
        """Render using ASCII Box

        :return: Multiline String
        """
        # determine length and guess type if not format provided
        columns = {
            i.name: [i.format, len(i.header) + 2 if i.is_key else len(i.header), i.show_none] for i in self._columns
        }
        logger.trace(*columns.items())
        rows = []
        for row in self._rows:
            row = row[:]
            for n, c in enumerate(columns.keys()):
                if isinstance(row[n], bool):
                    if not columns[c][0]:
                        columns[c][0] = "^"
                    columns[c][1] = max(columns[c][1], 5)
                    row[n] = "*True*" if row[n] else "*False*"
                elif isinstance(row[n], (int, float)):
                    if not columns[c][0]:
                        columns[c][0] = ">"
                    columns[c][1] = max(columns[c][1], len(str(row[n])))
                elif row[n] is None:
                    row[n] = "*None*" if columns[c][2] else ""
                    columns[c][1] = max(columns[c][1], len(row[n]))
                else:
                    columns[c][1] = max(columns[c][1], len(row[n]))
            rows.append(row)
        logger.trace(*columns.items())

        formatter = "║ {:" + "} │ {:".join([f'{columns[i][0] or "<"}{str(columns[i][1])}' for i in columns]) + "} ║"

        result = [
            "╔═" + "═╤═".join(["═" * columns[i][1] for i in columns.keys()]) + "═╗",
            formatter.format(*[f"*{i.header}*" if i.is_key else i.header for i in self._columns]),
            "╟─" + "─┼─".join(["─" * columns[i][1] for i in columns.keys()]) + "─╢",
        ]
        result.extend([formatter.format(*i) for i in rows])
        result.append("╚═" + "═╧═".join(["═" * columns[i][1] for i in columns.keys()]) + "═╝")

        return "\n".join(result)

    def markdown(self) -> str:
        """Render the provided data as a Markdown string.

        :return: Markdown string
        """
        # determine length and guess type if not format provided
        columns = {
            i.name: [
                i.format,
                len(i.header) + 4 if i.is_key else len(i.header),
                ":" if i.format in "<^" else "-",
                ":" if i.format and i.format in "^>" else "-",
                i.show_none,
            ]
            for i in self._columns
        }
        logger.trace(*columns.items())
        rows = []
        for row in self._rows:
            row = row[:]
            for n, c in enumerate(columns.keys()):
                if isinstance(row[n], bool):
                    if not columns[c][0]:
                        columns[c][0] = "^"
                        columns[c][2] = columns[c][3] = ":"
                    columns[c][1] = max(columns[c][1], 7)
                    row[n] = "&check;" if row[n] else ""  # convert to display_table chan
                elif isinstance(row[n], (int, float)):
                    if not columns[c][0]:
                        columns[c][0] = ">"
                        columns[c][3] = ":"
                    columns[c][1] = max(columns[c][1], len(str(row[n])))
                elif row[n] is None:
                    row[n] = "_None_" if columns[c][4] else ""
                    columns[c][1] = max(columns[c][1], len(row[n]))
                else:
                    columns[c][1] = max(columns[c][1], len(row[n]))
            rows.append(row)

        formatter = "| {:" + "} | {:".join([f'{columns[i][0] or "<"}{str(columns[i][1])}' for i in columns]) + "} |"

        result = [
            formatter.format(*[f"**{i.header}**" if i.is_key else i.header for i in self._columns]),
            "|" + "|".join([columns[i][2] + "-" * columns[i][1] + columns[i][3] for i in columns.keys()]) + "|",
        ]
        result.extend([formatter.format(*i) for i in rows])

        return "\n".join(result)



import os
import pathlib
import subprocess
import sys

from hakisto import Logger

logger = Logger("imuthes.make_link")


def make_link(
    target: pathlib.Path,
    link: pathlib.Path,
) -> None:
    """Create a link pointing to target.

    For Windoze, this will be a ``Junction``, for all other OS, a soft link
    :param target: Target path, this will be pointed to.
    :param link: Link path, this will be created.
    """
    logger.debug(f"{link} -> {target}")
    if sys.platform != "win32":
        logger.debug("This is not Windoze - using soft link")
        link = link.resolve()
        target = target.resolve()
        link.symlink_to(target)
    else:
        cmd = ["mklink", "/j", os.fsdecode(link), os.fsdecode(target)]
        logger.debug(f'calling {" ".join(cmd)}')
        proc = subprocess.run(cmd, shell=True, capture_output=True)
        if proc.returncode:
            raise OSError(proc.stderr.decode().strip())
